###################################################################################
## ExternalNews.py 
## (c)2001, Olivier DECKMYN
###################################################################################


"""This module purposes a thread-safe class to handle the management (load, 
parse, browse, update etc...) of a syndicated news channel respecting RDF or 
RSS format (XML based). 

Use provided ExternalNewsChannel class as this : 

Building :
>>> myChannel=ExternalNewsChannel('http://slashdot.org/slashdot.rdf', 'slashdot')

Browsing, Using : 
>>> for item in myChannel.getItems():
       print item['title'], item['link']

Updating : 
>>> myChannel.refresh()

Simple Printing : 
>>> print myChannel


Other interessting methods are : 
 - asDirtyHTML
 - getLastUpdateDateTime
"""

import os, time, string, sys
import urllib, urlparse
from xml.sax import saxlib, saxexts
from threading import Lock, Thread

###################################################################################
## Utility functions
###################################################################################

def clean_string(s):
    """Removes the too much spaces and crlf that might be in string"""
    return string.join(string.split(s))

###################################################################################
## Technical classes
###################################################################################

class MyRDFParserHandler(saxlib.HandlerBase):
    """This is a very technical class, providing delegated method to the 
    XML/RSS/RDF parser"""
    def __init__(self, channel):
	self.channel=channel      # Channel we are working on
        self._current_element=''  # 'item' , 'channel' or 'image'
	self._current_property='' # Current property for current element
	self._current_item=None   # Current news item, if current element is an item

    def startElement(self,ele,attr):
        """Method called by the parser when a element starts"""
        if ele in ['image', 'channel', 'item'] :
	    self._current_element=ele
            self._current_property=''
            if ele == 'item' : # Starting a new item
                self._current_item=self.channel._new_item() 
        elif self._current_element=="item" and ExternalNewsItem.DefaultProperties.has_key(ele) : 
	    # Working on a sub element of item (a property of item)
            self._current_property=ele
	elif self._current_element=="image" and ele in ["url"]:
	    # Working on url of image tag
            self._current_property=ele
	elif self._current_element=="channel" and ele in ["description", "title", "link"]:
	    # Working on url of image tag
            self._current_property=ele
	elif self._current_element!='':
	    # Working on url of image tag
            self._current_property=""	
	else:
	    pass
            #print "NO PROPERTY HANDLER FOR ", ele

    def endElement(self,ele):
        """Method called by the parser when a element ends"""
        if self._current_element=="item" and ExternalNewsItem.DefaultProperties.has_key(ele):
            self._current_property = ""
            #self._current_item.properties[ele] = clean_string(self._current_item.properties[ele])
	elif ele in ['image', 'channel', 'item']:
	    self._current_element=''
            if ele == 'item':
 	        self._current_item=None # end of current item
    
    def characters(self,ch,start,length):
        """Method called by the parser when reading text value (between element start and end tags)"""
	if self._current_property<>'':
	    s=ch[start:start+length]
  	    s=clean_string(s)
	    if self._current_element=="image" and self._current_property=="url":
	        self.channel.image_url=self.channel.image_url+s
            elif self._current_element=="channel" and self._current_property=="title":
	        self.channel.title=self.channel.title+s
            elif self._current_element=="channel" and self._current_property=="description":
	        self.channel.description=self.channel.description+s
	    elif self._current_element=="channel" and self._current_property=="link":
		self.channel.link=self.channel.link+s
	    elif self._current_element=='item' and self._current_property!='':
		self._current_item.properties[self._current_property]=self._current_item.properties[self._current_property]+s

###################################################################################
## Classes
###################################################################################

class ExternalNewsItem:

    DefaultProperties={
	  "title": "", 
	  "description" : "", 
	  "link" : "",
	}

    def __init__(self, channel):
        self.channel=channel
	self.properties={}
	self.properties.update(self.DefaultProperties)

    def __str__(self):
        result=[]
        for key in self.properties.keys():
	    result.append(key+"="+self.properties[key])
        return string.join(result,", ")

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

class ExternalNewsChannel:
    """
    Creates a Channel.
    see module documentation.
    Update is thread-safe.
    """
    def __init__(self, url, name="", min_delay=10, parse_now=1):
        self.url=url
	self.name=name

        # Protected variables
	self._min_delay=min_delay
	self._timestamp= 0
	self._lock=Lock()

	self.items=[] # list of news items

	# Properties
	self.image_url=''
	self.title=''
	self.description=''
	self.link=''

	if parse_now==1:	
	    self.download_and_parse()

    def refresh(self):
        """If our channel is out of date (time spent > freq parameter), the channel is destroyed, loaded and parsed again"""
        if time.time()-self._timestamp>=self._min_delay*60:
	    self.download_and_parse()
	    
    def download_and_parse(self):
        """Destroy, Download and Parses channel - Thread-Safe"""
	self.image_url=''
	self.title=''
	self.description=''
	self.link=''
	 
	self._lock.acquire()
	try:
            xmlp=saxexts.make_parser()  # Prepare parser
            dh=MyRDFParserHandler(self) # Prepare Document Handler
            xmlp.setDocumentHandler(dh) 
            self.clear_items()          # Clear actual items list
            xmlp.parse(self.url)        # Make the job (launches parsing + document handling)
            self._timestamp=time.time() # Store TimeStamp for refresh
	finally:
	    self._lock.release()


    def _new_item(self):
        """Make and return a new item. This news item is owned by the channel"""
        n=ExternalNewsItem(self)
	self.items.append(n)
	return n
 
    def clear_items(self):
        """Removes all items of the channel"""
        self.items=[]

    def asDirtyHTML(self):
        """Renders a channel in a dirty html way. Made to ease debug only."""
        str="<TABLE BORDER=0>\n"
        str=str + '<TH><A HREF="%s">%s</A></TH>\n' % ( self.link, self.name )
        str=str + "<TR><TD>\n  <TABLE>\n" 

        for i in self.getItems():
            str=str + '    <TR><TD><A HREF="%s">%s</A>' % ( i['link'],i['title'] )
            if i.has_key("description"):
                str=str + "<BR>%s" % ( i['description'])
            str=str + '</TD></TR>\n' 
        str=str+"</TABLE></TD></TR>\n</TABLE>\n\n"
        return str
        
    def __str__(self):
        """Very simple renderer, made for debug only."""
        result=[]
	##self.refresh()
	result.append('Name : '+self.name)
	result.append('Link : '+self.link)
	result.append('Title : '+self.title)
	result.append('Description : '+self.description)
	result.append('Image : '+self.image_url)
	result.append('Last updated on : '+self.getLastUpdateDateTime())
        for story in self.getItems():
	   result.append(str(story))
	return string.join(result,"\n")
    	
    def getLastUpdateDateTime(self):
        """Returns last update date time a string using YYYYMMDDHHMMSS"""
	t=time.localtime(self._timestamp)
        return time.strftime("%Y%m%d%H%M%S", t)

    def getItems(self):
        ##self.refresh()
	return self.items    	

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

class ExternalNewsManager:
    """This class handles the management of a list of channels. It is reponsible for 
    owning and automatically updating them."""
    def __init__(self):
        self.channels={} # Channels are stored in a dictionnary, index is channel name
	self.default_min_delay=10

    def auto_update(self, delay=60*5):    	
	self._updater=ExternalNewsManagerUpdater(self, delay)
	self._updater.start()

    def addChannel(self, url, name):
        """Create a new channel, given its name and url. Channel is appended to list of managed 
	   channels and returned as the result"""
        c=ExternalNewsChannel(url, name, self.default_min_delay, parse_now=0)
	self.channels[name]=c
	return c

    def getChannelCount(self):
        """Returns number of managed channels"""
	return len(self.channels.keys())    	
 
    def getChannelNames(self):
        """Returns a list with all channels names"""
        return self.channels.keys()

    def getChannel(self, name):
        """Return channel object given its name"""
        return self.channels[name]

    def Refresh(self):
        """Refreshes all channels."""
        for name in self.getChannelNames():
	    channel=self.getChannel(name)
	    channel.refresh()
	
    def __str__(self):
        return "<ExternalNewsManager with %d channel(s)>"% self.getChannelCount()
    	
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

class ExternalNewsManagerUpdater(Thread):
    """This class is a THREAD responsible for updating in the background all the channel 
    of a given channels manager every 'delay' seconds."""
  
    def __init__(self, manager, delay):
        Thread.__init__(self) 
	self._manager=manager
	self._delay=delay

    def run(self):
        while 1:
            self._manager.Refresh()
	    time.sleep(self._delay)

###################################################################################
## Tests
###################################################################################
def test_channel():
    s=ExternalNewsChannel('http://slashdot.org/slashdot.rdf', 'slashdot')
    print s

    s=ExternalNewsChannel('http://www.zope.org/SiteIndex/news.rss', 'zope')
    print s

    s=ExternalNewsChannel('http://www.freshmeat.net/backend/fm.rdf', 'freshmeat')
    print s


def test_manager():
    m=ExternalNewsManager()
    m.addChannel('http://slashdot.org/slashdot.rdf', 'slashdot')
    m.addChannel('http://www.zope.org/SiteIndex/news.rss', 'zope')
    m.addChannel('http://www.freshmeat.net/backend/fm.rdf', 'freshmeat')
    print m 

    print m.getChannel('zope')


def test_updater():
    m=ExternalNewsManager()
    m.addChannel('http://slashdot.org/slashdot.rdf', 'slashdot')
    m.addChannel('http://www.zope.org/SiteIndex/news.rss', 'zope')
    m.addChannel('http://www.freshmeat.net/backend/fm.rdf', 'freshmeat')
    m.addChannel('http://www.hardware.fr/backend/news.rdf', 'hardware')
    m.auto_update()
    while 1:
       time.sleep(1)
       print m.getChannel('zope')
       print m.getChannel('hardware')
   	

if __name__=="__main__":
    test_updater()
	